import csv
from io import StringIO
from typing import Dict, List, Tuple

import openpyxl

from hl7._config import INDEX_SEPARATOR


def csv_to_hl7(csv_content: str) -> str:
    return _to_hl7(_convert(csv_content))


def excel_to_hl7(file_path: str) -> str:
    return csv_to_hl7(
        "\n".join(
            ";".join(map(str, row[:3]))
            for row in openpyxl.load_workbook(file_path).active.iter_rows(
                values_only=True, min_row=2
            )
        )
    )


def _to_hl7(hl7_data: Dict[str, List[List[str]]]) -> str:
    message = ""
    for segment_name, segment_data in hl7_data.items():
        segment_value = "|".join(
            [
                "^".join(
                    [
                        "&".join([l3 or "" for l3 in l2 or []])
                        for l2 in l1 or []
                    ]
                )
                for l1 in segment_data
            ]
        )
        message += f"{segment_name}|{segment_value}\n"
    if message.startswith("MSH|||"):
        return message.replace("MSH|||", "MSH|")
    if message.startswith("MSH||"):
        return message.replace("MSH||", "MSH|")


def _hl7_index_split(index: str) -> Tuple[int, int, int]:
    indexes = index.split(INDEX_SEPARATOR)
    try:
        index_pipes = int(indexes[0]) - 1
    except IndexError:
        raise ValueError("Error")
    try:
        index_roofs = int(indexes[1]) - 1
    except IndexError:
        index_roofs = 0
    try:
        index_ands = int(indexes[2]) - 1
    except IndexError:
        index_ands = 0
    return index_pipes, index_roofs, index_ands


def _convert(content: str) -> Dict[str, List[List[str]]]:
    hl7_data: Dict = {}
    delimiters_to_try = [";", ","]

    for delimiter in delimiters_to_try:
        data = list(csv.reader(StringIO(content), delimiter=delimiter))

        if len(data[0]) >= 3:
            break

    for segment_name, index, value in map(lambda elem: elem[:3], data):
        index_l1, _, _ = _hl7_index_split(index)
        if segment_name in hl7_data:
            hl7_data[segment_name] = hl7_data[segment_name] + (
                [None] * (index_l1 + 1 - len(hl7_data[segment_name]))
            )
        else:
            hl7_data[segment_name] = [None] * (index_l1 + 1)

    for segment_name, index, value in map(lambda elem: elem[:3], data):
        index_l1, index_l2, _ = _hl7_index_split(index)

        l2_list: List | None = hl7_data[segment_name][index_l1]
        if l2_list is not None:
            hl7_data[segment_name][index_l1] = l2_list + (
                [None] * (index_l2 + 1 - len(l2_list))
            )
        else:
            hl7_data[segment_name][index_l1] = [None] * (index_l2 + 1)

    for segment_name, index, value in map(lambda elem: elem[:3], data):
        index_l1, index_l2, index_l3 = _hl7_index_split(index)
        l3_list: List | None = hl7_data[segment_name][index_l1][index_l2]
        if l3_list is not None:
            hl7_data[segment_name][index_l1][index_l2] = l3_list + (
                [None] * (index_l3 + 1 - len(l3_list))
            )
        else:
            hl7_data[segment_name][index_l1][index_l2] = [None] * (
                index_l3 + 1
            )

    for segment_name, index, value in map(lambda elem: elem[:3], data):
        index_l1, index_l2, index_l3 = _hl7_index_split(index)
        hl7_data[segment_name][index_l1][index_l2][index_l3] = value

    return hl7_data



import os

BASE_PATH = os.path.dirname(os.path.abspath(__file__))

CURRENT_HL7_VERSION = "2.5.1"
FIELD_SEPARATOR = "|"
INDEX_SEPARATOR = "."

ENCODING_CHARACTERS = (
    "^",
    "~",
    "\\",
    "&",
)



import json
import os
from io import BytesIO
from typing import Iterator, Tuple

import openpyxl

from ._config import (
    BASE_PATH,
    CURRENT_HL7_VERSION,
    ENCODING_CHARACTERS,
    FIELD_SEPARATOR,
)

_FIELD_SEPARATOR_FIELD = "<FIELD_SEPARATOR_FIELD>"
_ENCODING_CHARS_FIELD = "<ENCODING_CHARS_FIELD>"
_ENCODING_CHARACTERS_STR = "".join(ENCODING_CHARACTERS)

with open(
    os.path.join(BASE_PATH, "data", f"hl7v{CURRENT_HL7_VERSION}.json"), "r"
) as hl7_meta_data_file:
    HL7_META_DATA = json.loads(hl7_meta_data_file.read())


def hl7_to_csv(hl7_content: str) -> str:
    if f"MSH|{_ENCODING_CHARACTERS_STR}" not in hl7_content:
        raise ValueError("Not a valid HL7 message")

    return _prepare_csv_content(
        "\n".join(
            map(
                lambda csv_line: ";".join(csv_line),
                _convert(
                    hl7_content=_prepare_hl7_content(hl7_content=hl7_content)
                ),
            )
        )
    )


def hl7_to_excel(hl7_content):
    if f"MSH|{_ENCODING_CHARACTERS_STR}" not in hl7_content:
        raise ValueError("Not a valid HL7 message")

    output = BytesIO()
    workbook = openpyxl.Workbook()
    worksheet = workbook.active
    for segment, index, value, description in _convert(
        hl7_content=_prepare_hl7_content(hl7_content=hl7_content)
    ):
        if segment == "MSH" and value in [
            _FIELD_SEPARATOR_FIELD,
            _ENCODING_CHARS_FIELD,
        ]:
            value = _prepare_csv_content(value)
        worksheet.append((segment, index, value, description))

    workbook.save(output)
    output.seek(0)
    return output.read()


def _convert(hl7_content: str) -> Iterator[Tuple[str, str, str, str]]:
    for segment in hl7_content.split("\n"):
        segment = segment.strip()
        fields1 = segment.split(FIELD_SEPARATOR)
        segment_name = fields1[0]
        for index1, field in enumerate(fields1[1:], start=1):
            if not field:
                continue
            for index2, field2 in enumerate(field.split("^"), start=1):
                if not field2:
                    continue
                if index2 == 1:
                    description = _get_description(segment_name, int(index1))
                else:
                    description = _get_description(segment_name, int(index1), int(index2))
                if "&" in field2:
                    for index3, fields3 in enumerate(
                        field2.split("&"), start=1
                    ):
                        if index3 == 1:
                            description = _get_description(segment_name, int(index1), int(index2))
                        else:
                            description = _get_description(segment_name, int(index1), int(index2), int(index3))
                        yield segment_name, f"{index1}.{index2}.{index3}", fields3, description
                else:
                    yield segment_name, f"{index1}.{index2}", field2, description


def _prepare_hl7_content(hl7_content: str) -> str:
    return hl7_content.replace(
        f"MSH|{_ENCODING_CHARACTERS_STR}",
        f"MSH|{_FIELD_SEPARATOR_FIELD}|{_ENCODING_CHARS_FIELD}",
    )


def _prepare_csv_content(csv_content: str) -> str:
    return csv_content.replace(
        _ENCODING_CHARS_FIELD, _ENCODING_CHARACTERS_STR
    ).replace(_FIELD_SEPARATOR_FIELD, FIELD_SEPARATOR)


def _get_description(segment_name: str, index1: int, index2: int | None = None, index3: int | None = None) -> str:
    if index2 is not None:
        key_json = f"{segment_name}.{index1}.{index2}"
    else:
        key_json = f"{segment_name}.{index1}"

    if key_json in HL7_META_DATA:
        return HL7_META_DATA[key_json]["name"]
    return "Description not found"

